/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)41       1.25.1.2  src/avs/fs/mmfs/ts/kernext/gpl-linux/verdep.h, mmfs, avs_rcs2b 9/28/04 09:36:16 */
#ifndef _h_verdep
#define _h_verdep

#if (LINUX_KERNEL_VERSION >= 2060000)
/************************************
 * Linux 2.6 functional differences *
 ************************************/
# include <linux/buffer_head.h>
# include <linux/kdev_t.h>
# include <linux/namei.h>
# include <linux/mount.h>
# include <linux/page-flags.h>

# include <cxiTypes.h>
# include <cxiSystem.h>
# include <Logger-gpl.h>

  /* 2.4 kernel supports fh_to_dentry/dentry_to_fh callbacks */
  /* 2.6 has encode_fh and decode_fh */

# define GPFS_HANDLE_NFS_FH

  /* Retrieve a new empty inode from the vfs */
# define NEW_INODE(SUPER)            new_inode(SUPER)
 
  /* Uncache a directory entry.  Must hold dcache lock. */
# define DENTRY_DROP(DP)             __d_drop(DP)

  /* Check if directory is already uncached.  Must hold dcache lock. */
# define DENTRY_UNHASHED(DP)         d_unhashed(DP)

  /* Mark buffer in error and call iodone routine */
# define BUFHEAD_ERROR_IODONE(BH)  \
  {                                \
    clear_buffer_uptodate(BH);     \
    (BHP)->b_end_io((BH), 0);      \
  }

# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
  }

# define DAEMONIZE(A)                daemonize(A);
# define FOR_EACH_TASK(T)            for_each_process(T)
# define NEXT_TASK(T)                list_entry((T)->tasks.next, \
                                                 struct task_struct, tasks)
# define PROCESS_GROUP(T)            process_group(T)
# define REVOKE_LEASE(INODE,MODE)    break_lease(INODE,MODE)
# define KDEV_INT(KDEV)              ((unsigned int)(KDEV))
# define PAGE_UP_TO_DATE(PAGE)       PageUptodate(PAGE)
# define PAGE_UNLOCK(PAGE)           unlock_page(PAGE)
# define HAS_IOP_REVALIDATE          0  /* uses getattr inode op instead */
# define HAS_SOP_READ_INODE2         0  
# define SBLOCK_PRIVATE(SBP)         (SBP)->s_fs_info
# define SBLOCK_BDEVNAME(SBP,BNAME)  bdevname((SBP)->s_bdev,(BNAME))

# define DCACHE_DFLAGS_DISCONNECTED  DCACHE_DISCONNECTED
# define POSIX_UNBLOCK_LOCK(FLP)     posix_unblock_lock(NULL,(FLP))
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2))
# define SPIN_LOCK_IRQ(LOCK,FLAGS)   spin_lock_irqsave(LOCK,FLAGS)
# define SPIN_UNLOCK_IRQ(LOCK,FLAGS) spin_unlock_irqrestore(LOCK,FLAGS)

# define GET_INODETIME_SEC(ITIME)  (ITIME).tv_sec
# define CXITIME_TO_INODETIME(CXITIME, ITIME)   \
  {                                             \
    (ITIME).tv_sec = (CXITIME).tv_sec;          \
    (ITIME).tv_nsec = (CXITIME).tv_nsec;        \
  }
# define CXITIME_FROM_INODETIME(CXITIME, ITIME) \
  {                                             \
    (CXITIME).tv_sec = (ITIME).tv_sec;          \
    (CXITIME).tv_nsec = (ITIME).tv_nsec;        \
  }

#ifdef MODULE
  /* Module function interfaces for 2.6 kernel */
# define MY_INIT_FUNCTION()           static int __init myModuleInit(void)
# define MY_EXIT_FUNCTION()           static void __exit myModuleExit(void)
# define MY_MODULE_INCREMENT()        try_module_get(THIS_MODULE)
# define MY_MODULE_DECREMENT()        module_put(THIS_MODULE) 

# define MODULE_INCREMENT(MOD)        try_module_get(MOD)
# define MODULE_DECREMENT(MOD)        module_put(MOD)

# define DEFINE_MODULE_INIT()         module_init(myModuleInit)
# define DEFINE_MODULE_EXIT()         module_exit(myModuleExit)
#endif /* MODULE */

  /* pte functions */
# define PTE_OFFSET(PMDP,VADDR)       pte_offset_map((PMDP),(VADDR))
# define PTE_UNMAP(PTEP)              pte_unmap(PTEP);

  /* mmap functions */
# define ACQUIRE_MMAP_SEM(SEM)        down_write(SEM)
# define RELEASE_MMAP_SEM(SEM)        up_write(SEM)

# define FILEMAP_FDATASYNC(RC, M)     RC = filemap_fdatawrite(M)
# define FILEMAP_FDATAWAIT(RC, M)     RC = filemap_fdatawait(M)

# define MOD_PGPGIN(amount)           add_page_state(pgpgin, (amount))
# define MOD_PGPGOUT(amount)          add_page_state(pgpgout, (amount))

#else /* < 2060000 */
/************************************
 * Linux 2.4 functional differences *
 ********************************** */
# include <linux/locks.h>
# include <linux/kernel_stat.h>

# include <cxiTypes.h>
# include <cxiSystem.h>
# include <Logger-gpl.h>

# define GPFS_HANDLE_NFS_FH

#if defined(REDHAT_AS_LINUX) || defined(SUSE_LINUX) ||\
    (defined(REDHAT_LINUX) && LINUX_KERNEL_VERSION >= 2041814)
  /* Retrieve a new empty inode from the vfs */
# define NEW_INODE(SUPER)            new_inode(SUPER)
#else
# define NEW_INODE(SUPER)            get_empty_inode()
#endif

  /* Uncache a directory entry.  Assumes dcache lock is held */
# define DENTRY_DROP(DP)             list_del_init(&(DP)->d_hash)

  /* Check if directory is already uncached.  Must hold dcache lock. */
# define DENTRY_UNHASHED(DP)         list_empty(&(DP)->d_hash)

  /* Mark buffer in error and call iodone routine */
# define BUFHEAD_ERROR_IODONE(BH)    buffer_IO_error(BH)

#if LINUX_KERNEL_VERSION >= 2041800
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
  }
#else
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
  }
#endif

#if ((defined(REDHAT_LINUX) && (LINUX_DISTRIBUTION_LEVEL >= 90)) || \
     (defined(REDHAT_AS_LINUX) && (LINUX_KERNEL_VERSION >= 2042000)))
# define HAS_01_SCHED_THREADING
#endif

#ifdef HAS_01_SCHED_THREADING
# define NEXT_TASK(T)                list_entry((T)->tasks.next, \
                                                struct task_struct, tasks)
# define FOR_EACH_TASK(T)            for ((T) = &init_task;        \
                                          ((T) = NEXT_TASK(T)) != &init_task; )
#else
# define FOR_EACH_TASK(T)            for_each_task(T)
# define NEXT_TASK(T)                (T)->next_task
#endif
# define PROCESS_GROUP(T)            (T)->pgrp

# define DAEMONIZE(A)                do { daemonize(); reparent_to_init(); } while (0)
# define REVOKE_LEASE(INODE,MODE)    get_lease(INODE,MODE)
# define KDEV_INT(KDEV)              (KDEV)
# define PAGE_UP_TO_DATE(PAGE)       Page_Uptodate(PAGE)
# define PAGE_UNLOCK(PAGE)           UnlockPage(PAGE)
# define HAS_SOP_READ_INODE2         1 /* read_inode2 super block op present */
# define HAS_IOP_REVALIDATE          1 /* revalidate inode op present */
# define SBLOCK_PRIVATE(SBP)         (SBP)->u.generic_sbp
# define SBLOCK_BDEVNAME(SBP,BNAME)  bdevname((SBP)->s_dev)
# define BDEVNAME_SIZE               0 /* char array unneeded on 2.4 */

# define DCACHE_DFLAGS_DISCONNECTED  DCACHE_NFSD_DISCONNECTED
# define POSIX_UNBLOCK_LOCK(FLP)     posix_unblock_lock(FLP)
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2),0)
# define SPIN_LOCK_IRQ(LOCK,FLAGS)   spin_lock_irqsave(LOCK,FLAGS)
# define SPIN_UNLOCK_IRQ(LOCK,FLAGS) spin_unlock_irqrestore(LOCK,FLAGS)

# define GET_INODETIME_SEC(ITIME)  (ITIME)
# define CXITIME_TO_INODETIME(CXITIME, ITIME)   \
  {                                             \
    (ITIME) = (CXITIME).tv_sec;                 \
  }
# define CXITIME_FROM_INODETIME(CXITIME, ITIME) \
  {                                             \
    (CXITIME).tv_sec = (ITIME);                 \
    (CXITIME).tv_nsec = 0;                      \
  }

#ifdef MODULE
  /* Module function interfaces for 2.4 kernel */
# define MY_INIT_FUNCTION()           int init_module(void)
# define MY_EXIT_FUNCTION()           void cleanup_module(void)
# define MY_MODULE_INCREMENT()        MOD_INC_USE_COUNT
# define MY_MODULE_DECREMENT()        if (MOD_IN_USE) \
                                        MOD_DEC_USE_COUNT

# define MODULE_INCREMENT(MOD)        __MOD_INC_USE_COUNT(MOD)
# define MODULE_DECREMENT(MOD)        if (__MOD_IN_USE(MOD)) \
                                        __MOD_DEC_USE_COUNT(MOD)
# define DEFINE_MODULE_INIT()
# define DEFINE_MODULE_EXIT()
#else
   /* No ops if not building a module */
# define MY_INIT_FUNCTION()          int myInit(void)
# define MY_EXIT_FUNCTION()          void myExit(void)
# define MY_MODULE_INCREMENT()
# define MY_MODULE_DECREMENT()

# define MODULE_INCREMENT(MOD)
# define MODULE_DECREMENT(MOD)

# define DEFINE_MODULE_INIT()        module_init(myInit)
# define DEFINE_MODULE_EXIT()        module_exit(myExit)
#endif /* !MODULE */

  /* pte functions */
#if defined(REDHAT_AS_LINUX) || \
    (defined(REDHAT_LINUX) && (LINUX_KERNEL_VERSION >= 2042006))
# define PTE_OFFSET(PMDP,VADDR)  pte_offset_map((PMDP),(VADDR))
# define PTE_UNMAP(PTEP)         pte_unmap(PTEP);
#else
# define PTE_OFFSET(PMDP,VADDR)  pte_offset(PMDP, VADDR)
# if defined(SUSE_LINUX) && defined(GPFS_ARCH_I386)
#   define PTE_UNMAP(PTEP)       pte_kunmap(PTEP);
# else
#   define PTE_UNMAP(PTEP)       /* nothing */
# endif
#endif /* pte functions */

#if LINUX_KERNEL_VERSION >= 2040312 || GPFS_ARCH_POWER
# define ACQUIRE_MMAP_SEM(SEM)   down_write(SEM)
# define RELEASE_MMAP_SEM(SEM)   up_write(SEM)

#else
# define ACQUIRE_MMAP_SEM(SEM)   down(SEM)
# define RELEASE_MMAP_SEM(SEM)   up(SEM)
#endif

#if LINUX_KERNEL_VERSION >= 2041800
  /* Earlier kernels do not have a return code */
# define FILEMAP_FDATASYNC(RC, M) RC = filemap_fdatasync(M)
# define FILEMAP_FDATAWAIT(RC, M) RC = filemap_fdatawait(M)
#else
# define FILEMAP_FDATASYNC(RC, M) do { filemap_fdatasync(M); RC = 0; } while (0)
# define FILEMAP_FDATAWAIT(RC, M) do { filemap_fdatawait(M); RC = 0; } while (0)
#endif

#if (defined(REDHAT_AS_LINUX)) && \
    (LINUX_KERNEL_VERSION >= 2042101)
# define MOD_PGPGIN(amount)  kstat_percpu[smp_processor_id()].pgpgin += amount
# define MOD_PGPGOUT(amount) kstat_percpu[smp_processor_id()].pgpgout+= amount
#else
# define MOD_PGPGIN(amount)  kstat.pgpgin += amount
# define MOD_PGPGOUT(amount) kstat.pgpgout += amount
#endif

#endif /* 2.4 kernel */
#endif /* _h_verdep */
